from w.tests.builders.abstract_test_builder import AbstractTestBuilder
from w.tests.fixtures.datasets.builders.factory_boy import (
    ExampleWithAutoNowFactory,
)


class BuilderWithAutoNow(AbstractTestBuilder):
    _factories = {"example": ExampleWithAutoNowFactory}
    _main_factory_key = "example"



import unicodedata

from w.services.abstract_service import AbstractService


class StringService(AbstractService):
    @classmethod
    def strip_accents(cls, txt) -> str:
        """
        Replace the characters with accents in txt by their normal form.

        Args:
            txt(str): text to clean

        Returns:
            str: cleaned text
        """
        return "".join(
            (
                c
                for c in unicodedata.normalize("NFD", txt)
                if unicodedata.category(c) != "Mn"
            )
        )

    @classmethod
    def clean(cls, txt, options=None) -> str:
        """
        Clean the provided text by :
            - applying all the replacements specified in options, if any
            - removing space from both left and right,
            - converting all uppercase characters into lower ones
            - removing all accents

        Args:
            txt(str): text to clean
            options(dict): substring replacements :
                {"old_substring": "new_substring", ...}

        Returns:
            str: cleaned text
        """
        cleaned_txt = txt
        if options is not None:
            for old, new in options.items():
                cleaned_txt = cleaned_txt.replace(old, new)
        return cls.strip_accents(cleaned_txt.strip().lower())



from pathlib import Path

import openpyxl
from openpyxl.utils.exceptions import InvalidFileException

from w.services.abstract_service import AbstractService
from w.services.technical.filesystem_service import FilesystemService


class ExcelService(AbstractService):
    @classmethod
    def is_excel(cls, filename):
        """ Check if filename is excel (from extension) """
        return FilesystemService.has_suffix(filename, [".xls", ".xlsx"])

    @classmethod
    def check_is_excel(cls, filename) -> Path:
        """
        Check if filename is excel (from extension) or raise RuntimeError
        """
        if cls.is_excel(filename):
            return filename
        raise RuntimeError(f"{Path(filename).name} is not a valid excel file")

    @classmethod
    def open_workbook(cls, filename, **options):
        """
        Open filename is valid excel (check exists too)

        Args:
            filename(str): excel full path filename
            **options:
             - read_only(bool): default False, optimised for reading, content cannot be
               edited
             - keep_vba(bool): default False, preserve vba content
             - data_only(bool): default False, controls whether cells with formulae have
               either the formula (default) or the value stored the last time Excel read
               the sheet
            - keep_links: default True, whether links to external workbooks should be
              preserved.

        Returns:
            Workbook

        Raises:
            RuntimeError: if filename does not exists or is not valid excel
        """
        FilesystemService.check_file_exists(filename)
        cls.check_is_excel(filename)
        try:
            return openpyxl.load_workbook(filename, **options)
        except InvalidFileException:
            raise RuntimeError(f"{Path(filename).name} is not a valid excel file")

    @classmethod
    def load(cls, excel_filename, mapping_columns, sheet_name=None) -> list:
        """
        Load active or specific sheet of excel file

        Args:
            excel_filename(str|Path): excel file
            mapping_columns(dict): csv header columns mapping to wanted attributes
            sheet_name(str) : sheet name (default load active sheet)

        Returns:
            list: [{"<mapping name>": < row col value>, ...}, ...]

        Raises:
            RuntimeError :
                - filename is not csv
                - filename does not exists
                - incorrect or missing header
        """
        wb = cls.open_workbook(excel_filename, read_only=True, data_only=True)

        if sheet_name is not None and sheet_name not in wb.sheetnames:
            raise RuntimeError(f"{sheet_name} is not a valid sheet name")

        sheet = wb[sheet_name] if sheet_name is not None else wb.active

        rows = sheet.iter_rows()
        headers = [c.value.strip() for c in next(rows)]
        required_headers = list(mapping_columns.keys())
        if required_headers != headers:
            raise RuntimeError(
                f"incorrect or missing header, expected '{', '.join(required_headers)}'"
                f" got '{', '.join(headers)}'"
            )
        return [
            {mapping_columns[headers[i]]: cell.value for i, cell in enumerate(row)}
            for row in rows
        ]



from w.django.tests.django_testcase import DjangoTestCase
from w.tests.fixtures.datasets.django_app import models
from w.tests.fixtures.datasets.services.model_services import ExampleService


class TestAbstractModelService(DjangoTestCase):
    @classmethod
    def setUpTestData(cls):
        pass

    """
    create
    """

    def test_create_with_success_return_instance(self):
        """ Ensure create succeed """
        actual = ExampleService.create(attribute_one="one", attribute_two="two")
        assert isinstance(actual, models.Example)
        assert actual.attribute_one == "one"
        assert actual.attribute_two == "two"

    # todofsc: finir les tests



from w.tests.builders.abstract_test_builder import AbstractTestBuilder
from w.tests.fixtures.datasets.builders.factory_boy import SimpleExampleFactory


class SimpleBuilder(AbstractTestBuilder):
    _factories = {"example": SimpleExampleFactory}
    _main_factory_key = "example"



from django.contrib.auth.models import Permission
from django.db import connection
from django.test import RequestFactory, TestCase

from w.django import utils
from w.tests.mixins.testcase_mixin import TestCaseMixin


class DjangoTestCase(TestCaseMixin, TestCase):
    """ For testing in Django Context """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.reset_all_db_sequences()

    @classmethod
    def reset_all_db_sequences(cls):
        if connection.vendor != "postgresql":
            return None

        sql = """
        SELECT seqclass.relname     AS sequence_name,
               depclass.relname     AS table_name
        FROM   pg_class AS seqclass
               JOIN pg_sequence AS seq
                 ON ( seq.seqrelid = seqclass.relfilenode )
               JOIN pg_depend AS dep
                 ON ( seq.seqrelid = dep.objid )
               JOIN pg_class AS depclass
                 ON ( dep.refobjid = depclass.relfilenode )
        WHERE seqclass.relkind = 'S'
        """
        with connection.cursor() as c:
            c.execute(sql)
            results = c.fetchall()

            for sequence in results:
                sql = f"SELECT MAX(id) FROM {sequence[1]}"
                c.execute(sql)
                max_id = c.fetchall()[0][0]
                if max_id is None:
                    max_id = 0
                sql = f"ALTER SEQUENCE {sequence[0]} RESTART {max_id + 1};"
                c.execute(sql)

    @staticmethod
    def get_request(url, data=None, method="get", **extra):
        request = RequestFactory()

        if method == "get":
            return request.get(url, data, **extra)
        return request.post(url, data, **extra)

    @staticmethod
    def add_user_permission(user, permission):
        """
        Add permission to user
        """
        permission = Permission.objects.get(codename=permission)
        user.user_permissions.add(permission)

    @staticmethod
    def reverse(url_name, params=None, query_params=None):
        """ "
        Url reverse with query string management

        Usage:
            self.reverse(
                'app.views.my_view',
                params={'pk': 123},
                query_kwargs={'key':'value', 'k2': 'v2'}
            )
        """
        return utils.reverse(
            viewname=url_name, kwargs=params, query_kwargs=query_params
        )



import datetime

import arrow
import pytest

from w.services.technical.date_service import DateService
from w.tests.helpers import date_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestDateService(TestCaseMixin):
    @staticmethod
    def setup_method():
        DateService.clear()

    """
    previous_weekday
    """

    def test_previous_weekday_with_date_equals_seeking_day(self):
        actual = DateService.previous_weekday(arrow.get("2017-10-04"), DateService.WED)
        expected = arrow.get("2017-10-04")
        assert expected == actual

    def test_previous_weekday_with_date_after_seeking_day(self):
        actual = DateService.previous_weekday(arrow.get("2017-10-06"), DateService.WED)
        expected = arrow.get("2017-10-04")
        assert expected == actual

    def test_previous_weekday_with_date_before_seeking_day(self):
        actual = DateService.previous_weekday(arrow.get("2017-10-03"), DateService.WED)
        expected = arrow.get("2017-09-27")
        assert expected == actual

    """
    next_weekday
    """

    def test_next_weekday_with_date_equals_seeking_day(self):
        actual = DateService.next_weekday(arrow.get("2017-10-04"), DateService.WED)
        expected = arrow.get("2017-10-04")
        assert expected == actual

    def test_next_weekday_with_date_after_seeking_day(self):
        actual = DateService.next_weekday(arrow.get("2017-10-06"), DateService.WED)
        expected = arrow.get("2017-10-11")
        assert expected == actual

    def test_next_weekday_with_date_before_seeking_day(self):
        actual = DateService.next_weekday(arrow.get("2017-10-03"), DateService.WED)
        expected = arrow.get("2017-10-04")
        assert expected == actual

    """
    get_weekday_date
    """

    def test_get_weekday_date_with_seeking_day_same_as_date_return_date(self):
        """
        Ensure return will be the same as date if seeking day is same as day date
        and use datetime.datetime as date
        """
        date = datetime.date(2018, 4, 25)
        assert arrow.get("2018-04-25") == DateService.get_weekday_date(
            date, DateService.WED
        )

    def test_get_weekday_date_with_seeking_day_before_return_date(self):
        """
        Ensure get date succeed with seeking day before date and use Arrow as date
        """
        assert arrow.get("2018-04-30") == DateService.get_weekday_date(
            arrow.get("2018-05-02"), DateService.MON
        )

    def test_get_weekday_date_with_seeking_day_after_date_return_date(self):
        """
        Ensure get date succeed with seeking day after date and use datetime.date
        as date
        """
        date = datetime.datetime(2018, 4, 25)
        assert arrow.get("2018-04-28") == DateService.get_weekday_date(
            date, DateService.SAT
        )

    """
    to_mysql_date
    """

    def test_to_mysql_date_with_success_return_str(self):
        expected = "2018-05-01"
        assert (
            datetime.datetime.now().strftime("%Y-%m-%d") == DateService.to_mysql_date()
        )
        assert expected == DateService.to_mysql_date(datetime.datetime(2018, 5, 1))
        assert expected == DateService.to_mysql_date(datetime.date(2018, 5, 1))
        assert expected == DateService.to_mysql_date(arrow.get("2018-05-01"))

    """
    to_mysql_datetime
    """

    def test_to_mysql_datetime_with_success_return_str(self):
        expected = "2018-05-01 15:18:36.224469"

        with date_test_helper.today_is(expected):
            assert expected == DateService.to_mysql_datetime()

        assert expected == DateService.to_mysql_datetime(
            datetime.datetime(2018, 5, 1, 15, 18, 36, 224469)
        )
        assert expected == DateService.to_mysql_datetime(
            arrow.get("2018-05-01 15:18:36.224469")
        )

    def test_to_mysql_datetime_with_timezone_return_str(self):
        DateService.set_timezone("Europe/Paris")

        expected = "2018-05-01 15:18:36.224469"

        with date_test_helper.today_is("2018-05-01 13:18:36.224469"):
            assert expected == DateService.to_mysql_datetime()

        assert expected == DateService.to_mysql_datetime(
            datetime.datetime(2018, 5, 1, 13, 18, 36, 224469)
        )
        assert expected == DateService.to_mysql_datetime(
            arrow.get("2018-05-01 13:18:36.224469")
        )

    """
    week_number
    """

    def test_week_number_success_return_int(self):
        assert 18 == DateService.week_number(datetime.datetime(2018, 5, 1))
        assert 18 == DateService.week_number(datetime.date(2018, 5, 1))
        assert 18 == DateService.week_number(arrow.get("2018-05-01"))
        assert (
            DateService.week_number(datetime.date.today()) == DateService.week_number()
        )

    """
    to_date
    """

    def test_to_date_with_success_return_date(self):
        assert isinstance(DateService.to_date(), datetime.date)
        assert isinstance(
            DateService.to_date("2019-03-18T11:16:34.786+01:00"), datetime.date
        )

    """
    is_between
    """

    def test_is_between_with_end_before_start_raise_runtime_error(self):
        with pytest.raises(RuntimeError):
            DateService.is_between("2018-03-05", "2018-03-05", "2018-02-07")

    def test_is_between_with_date_not_between_return_false(self):
        assert not DateService.is_between(DateService.now(), "2018-02-05", "2018-02-06")

    def test_is_between_with_success_return_true(self):
        assert DateService.is_between("2018-02-05", "2018-02-05", "2018-02-06")
        assert DateService.is_between("2018-02-06", "2018-02-05", "2018-02-06")
        assert DateService.is_between("2018-02-06", "2018-02-05", "2018-02-07")

    """
    range_day
    """

    def test_range_day_with_end_before_start_raise_runtime_error(self):
        with pytest.raises(RuntimeError):
            DateService.range_day("2018-03-05", "2018-02-07")

    def test_range_day_with_success_return_list(self):
        days = DateService.range_day("2018-02-05", "2018-02-07")
        self.assert_equals_resultset([DateService.to_mysql_date(d) for d in days])

    def test_range_day_with_only_week_end_return_list(self):
        days = DateService.range_day("2019-04-04", "2019-04-09", only_week_end=True)
        self.assert_equals_resultset([DateService.to_mysql_date(d) for d in days])

    """
    is_week_end
    """

    def test_is_week_end_with_date_outside_week_end_return_false(self):
        """ Ensure date outside week end return False """
        assert not DateService.is_week_end("2019-04-30")

    def test_is_week_end_with_date_inside_week_end_return_true(self):
        """ Ensure date outside week end return False """
        assert DateService.is_week_end("2019-04-27")
        assert DateService.is_week_end("2019-04-28")

        with date_test_helper.today_is("2019-04-28"):
            assert DateService.is_week_end()

    """
    add_days
    """

    def test_add_days_with_success_return_date(self):
        """Ensure adding nb days to a date succeed"""
        assert "2019-05-13" == DateService.add_days("2019-04-29", 14).format(
            "YYYY-MM-DD"
        )

    def test_add_days_with_excluding_we_return_date(self):
        """Ensure adding nb days to a date succeed"""
        assert "2019-05-17" == DateService.add_days(
            "2019-04-29", 14, exclude_we=True
        ).format("YYYY-MM-DD")

    """
    nb_days
    """

    def test_nb_days_with_success_return_int(self):
        """Ensure nb days succeed"""
        assert 5 == DateService.nb_days("2019-03-01", "2019-03-06")
        assert -3 == DateService.nb_days("2019-03-09", "2019-03-06")

    def test_nb_days_with_we_excluded_return_int(self):
        """Ensure nb days succeed"""
        assert 4 == DateService.nb_days("2019-03-01", "2019-03-06", True)
        assert -6 == DateService.nb_days("2019-03-11", "2019-03-04", True)



import json
import logging
import re

logger = logging.getLogger(__name__)


class RequestResponse:
    """Request Response Model"""

    def __init__(self, response=None, **kwargs):
        """
        Args:
            response (requests.Response)
        """
        self._redirect_location = None
        if response is not None:
            self.url = response.url
            try:
                self._orig_content = response.content.decode()
                self.is_decoded = True
            except UnicodeDecodeError:
                self._orig_content = response.content
                self.is_decoded = False
            logger.info(f"Response status={response.status_code} to {self.url}")
            logger.debug(f"Response content={self._orig_content}")
            if response.ok:
                if self.is_decoded:
                    try:
                        self._content = json.loads(response.content)
                    except ValueError:
                        self._content = self._orig_content
                else:
                    self._content = self._orig_content
            else:
                self._content = (
                    self._orig_content if not response.reason else response.reason
                )

            self._success = response.ok
            self._status_code = response.status_code
            self._headers = response.headers
            if int(self._status_code) == 302:
                self._redirect_location = response.headers["Location"]
                logger.info(f"Redirect location={self._redirect_location}")
        else:
            # only for testing purpose
            self.url = kwargs.get("url", "http://dummy.com/")
            self._success = kwargs.get("success", None)
            self._status_code = kwargs.get("status_code", None)
            self._headers = kwargs.get("headers", None)
            self._orig_content = kwargs.get("orig_content", None)
            self.is_decoded = kwargs.get("is_decoded", True)
            if int(self._status_code) == 302:
                self._redirect_location = kwargs["headers"]["Location"]
            content = kwargs.get("content", None)
            try:
                self._content = json.loads(content)
            except ValueError:
                self._content = content

    @property
    def content(self):
        """dict: response content"""
        return self._content

    @property
    def success(self):
        """bool: request succeeded ?"""
        return self._success

    @property
    def status_code(self):
        """int: status code"""
        return self._status_code

    @property
    def orig_content(self):
        """original response content"""
        return self._orig_content

    @property
    def redirect_location(self):
        """ redirect location"""
        return self._redirect_location

    @property
    def is_redirect(self):
        """ Check if reponse is redirect """
        return self._redirect_location is not None

    def get_header(self, name):
        """ Get header name value """
        if name in self._headers:
            return self._headers[name]
        return None

    def get_content_filename(self):
        """ Get filename if exists """
        content_disposition = self.get_header("content-disposition")
        if content_disposition:
            fname = re.findall('filename="(.+)"', content_disposition)
            if len(fname) == 0:
                return None
            return fname[0]

        # parsing url to get the filename
        if not self.is_decoded and self.url.find("/"):
            return self.url.rsplit("/", 1)[1]
        return None

    def as_attachment(self):
        """ check if header Content-Disposition header specify as attachment """
        content_disposition = self.get_header("content-disposition")
        return content_disposition is not None and "attachment" in content_disposition

    def __str__(self) -> str:
        if self.success:
            return f"status code: {self.status_code}"
        return f"{self.status_code} - {self.content}"



# Works with drf.backends.jwt_authenticate and djangrestframeword_simplejwt
from django.conf import settings
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework_api_key.permissions import HasAPIKey
from rest_framework_simplejwt.exceptions import TokenError, InvalidToken
from rest_framework_simplejwt.serializers import TokenRefreshSerializer

from w.drf import viewsets


class JwtViewset(viewsets.ViewSet):
    serializers = {"default": TokenRefreshSerializer}
    permission_classes = (HasAPIKey,)

    @action(methods=["post"], detail=False)
    def refresh(self, request):
        """
        Refresh token via HttpOnly Cookie

        add to urls.py:
            router.register(r"auth", JwtViewset, basename="jwt_auth")
        """
        cookie_name = getattr(settings, "JWT_AUTH_COOKIE", None)
        if not cookie_name:
            raise RuntimeError("Refresh needs JWT_AUTH_COOKIE")

        serializer = self.get_serializer(
            data={"refresh": request.COOKIES.get(cookie_name)}
        )
        try:
            serializer.is_valid(raise_exception=True)
        except TokenError as e:
            raise InvalidToken(e.args[0])

        return Response(serializer.validated_data, status=status.HTTP_200_OK)



from rest_framework import serializers


class DrfValidator(serializers.Serializer):
    def update(self, instance, validated_data):
        raise RuntimeError("Forbidden, only for validation")

    def create(self, validated_data):
        raise RuntimeError("Forbidden, only for validation")



import sys
from abc import ABCMeta

from django.core.management import BaseCommand


class WCommand(BaseCommand, metaclass=ABCMeta):
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"

    def _write(self, msg, no_ending, stdout=None):
        if stdout is None:
            stdout = self.stdout

        if no_ending:
            stdout.write(msg, ending="")
        else:
            stdout.write(msg)

    def info(self, msg, no_ending=False):
        self._write(self.style.MIGRATE_HEADING(msg), no_ending)
        # force stdout print
        sys.stdout.flush()

    def success(self, msg, no_ending=False):
        self._write(self.style.SUCCESS(msg), no_ending)

    def warning(self, msg, no_ending=False):
        self._write(self.style.WARNING(msg), no_ending)

    def error(self, msg, no_ending=False):
        self._write(self.style.ERROR(msg), no_ending, stdout=self.stderr)



from io import StringIO

import yaml

from w.services.abstract_service import AbstractService
from w.services.technical.filesystem_service import FilesystemService


class YamlService(AbstractService):
    @staticmethod
    def dump_to_mem_file(data):
        """
        Convert a Yaml python object to in-memory yaml file
        """
        return StringIO(yaml.safe_dump(data, default_flow_style=False, sort_keys=False))

    @staticmethod
    def dump_to_file(data, filename):
        """
        Dump data to filename
        """
        with open(filename, "w") as f:
            yaml.safe_dump(data, f, default_flow_style=False, sort_keys=False)

    @staticmethod
    def load_from_file(filename):
        """
        Parse the YAML document from filename
        and produce the corresponding Python object.

        Args:
            filename(str):

        Returns:
            dict
        """
        FilesystemService.check_file_exists(filename)
        with open(filename) as f:
            return yaml.safe_load(f)

    @classmethod
    def check_is_valid(cls, yaml_stream):
        """
        Check is valid Yaml
        Args:
            yaml_stream():

        Returns:
            Dict: Yaml converted to dict

        Raises:
            RuntimeError
        """
        is_valid, result = cls.is_valid(yaml_stream)
        if is_valid:
            return result
        raise RuntimeError(result["user_message"])

    @staticmethod
    def is_valid(yaml_stream):
        """
        Check yaml is valid
        Args:
            yaml_stream:

        Returns:
            (Bool, result): yaml is valid ?,
                if is valid result = Yaml converted to dict
                else result = error as :
                {
                    "line": <line>,
                    "col": <col>,
                    "error": <error message>,
                    "user_message": <displayable message>
                }
        """
        try:
            return True, yaml.safe_load(yaml_stream)
        except yaml.YAMLError as e:
            # noinspection PyUnresolvedReferences
            err = {
                "line": e.problem_mark.line + 1,
                "col": e.problem_mark.column + 1,
                "error": e.problem,
                "user_message": (
                    f"Line {e.problem_mark.line+1}, "
                    f"column {e.problem_mark.column+1}: {e.problem}"
                ),
            }
            return False, err
        except ValueError as e:
            err = {"user_message": f"Date format error in rulebook: {e}"}
            return False, err



from requests import Session


class RequestSession:
    def __init__(self, base_url, **options):
        self.request = Session()
        self.base_url = base_url
        self.timeout = options["timeout"] if "timeout" in options else None
        auth = options.pop("auth", None)
        headers = options.pop("headers", None)
        cookies = options.pop("cookies", None)

        if auth:
            self.request.auth = auth
        if headers:
            self.request.headers = headers
        if cookies:
            self.request.cookies = cookies

    def render_url(self, relative_url):
        return f"{self.base_url}{relative_url}"

    def get_cookies(self):
        return self.request.cookies.get_dict()

    def __str__(self):
        return f"session on {self.base_url}"



from w.drf.viewsets import ModelServiceViewSet


class ListModelServiceMixin:
    def list(self: ModelServiceViewSet, request, *args, **kwargs):
        return self.get_list_response(self.service.list())


class RetrieveModelServiceMixin:
    def retrieve(self: ModelServiceViewSet, request, *args, **kwargs):
        return self.get_retrieve_response()


class CreateModelServiceMixin:
    def create(self: ModelServiceViewSet, request, *args, **kwargs):
        instance = self.service.create(**self.check_is_valid())
        return self.get_create_response(instance)


class UpdateModelServiceMixin:
    def update(self: ModelServiceViewSet, request, *args, **kwargs):
        instance = self.service.check_by_pk(self.kwargs["pk"])
        return self.get_update_response(
            self.service.update(instance, **self.check_is_valid())
        )


class DeleteModelServiceMixin:
    def destroy(self: ModelServiceViewSet, request, *args, **kwargs):
        self.service.delete_by_pk(self.kwargs["pk"])
        return self.get_delete_response()



from contextlib import contextmanager
from unittest.mock import patch

from w.services.technical.json_service import JsonService
from w.services.technical.filesystem_service import FilesystemService
from w.services.technical.models.request_response import RequestResponse


@contextmanager
def request_failure(response: dict, method="get"):
    """
    Request failure
    Args:
        response (dict): response params :
            {
                "url": <optional:url>
                "status_code": <optional:status code default 200>,
                "headers": <optional:list[dict]|dict: headers or list of headers>,
                "content": <optional:str|dict: content to use for response>,
                "json_file": <optional:json file or list json files for response>,
                "file": <optional:file for response>,
                "success": <bool: default True>
            }
        method(str): get|post|patch|put|del
    """
    default_response = {"status_code": 400, "success": False, "reason": "Bad Request"}
    response = {**default_response, **response}
    params = _get_patch_params(response, method)
    with patch(**params) as m:
        yield m


@contextmanager
def request_success(response: dict, method="get"):
    """
    Request success
    Args:
        response (dict): response params :
            {
                "url": <optional:url>
                "status_code": <optional:status code default 200>,
                "headers": <optional:list[dict]|dict: headers or list of headers>,
                "content": <optional:str|dict: content to use for response>,
                "json_file": <optional:json file or list json files for response>,
                "file": <optional:file for response>,
                "success": <bool: default True>
            }
        method(str): get|post|patch|put|del
    """
    default_response = {"status_code": 200}
    response = {**default_response, **response}

    with patch(**_get_patch_params(response, method)) as m:
        yield m


@contextmanager
def mock_request(responses, method="get"):
    """
    Request success
    Args:
        responses (list): list of responses
        method(str): get|post|patch|put|del
    """
    params = {
        "target": f"w.services.technical.request_service.RequestService.{method}",
        "side_effect": responses,
    }
    with patch(**params) as m:
        yield m


def get_response(**params):
    """
    Get request response

    Args:
        params (**): response params :
            {
                "url": <optional:url>
                "status_code": <optional:status code default 200>,
                "headers": <optional:list[dict]|dict: headers or list of headers>,
                "content": <optional:str|dict: content to use for response>,
                "json_file": <optional:str|list:json file or json files list for response>,
                "file": <optional:file for response>,
                "success": <bool: default True>
            }

    Returns:
        list(RequestResponse): list of response
    """
    default_params = {
        "content": "some content",
        "status_code": 200,
        "headers": {},
        "success": True,
    }
    params = {**default_params, **params}

    if "file" in params:
        params["content"] = FilesystemService.read_file(params["file"])

    if "content" in params and isinstance(params["content"], dict):
        params["content"] = JsonService.dump(params["content"])

    if "json_file" not in params:
        return RequestResponse(**params)

    if not isinstance(params["json_file"], list):
        params["json_file"] = (params["json_file"],)

    result = []
    for json_file in params["json_file"]:
        with open(json_file) as f:
            content = f.read()
            if "reason" not in params:
                params["content"] = content
            else:
                params["content"] = params["reason"]
                params["orig_content"] = content
            result.append(RequestResponse(**params))
    return result


def get_401_response(**params):
    default_params = {
        "return_content": "Unauhtorized",
        "status_code": 401,
        "success": False,
    }
    params = {**default_params, **params}
    return get_response(**params)


def get_400_response(**params):
    default_params = {"status_code": 400, "success": False, "reason": "Bad Request"}
    params = {**default_params, **params}
    return get_response(**params)


def _get_patch_params(response, method):
    response = get_response(**response)
    if not isinstance(response, list):
        response = [response]
    return {
        "target": f"w.services.technical.request_service.RequestService.{method}",
        "side_effect": response,
    }



from w.serializers import serializer


class RequestResponseSerializer(serializer.SerpySerializer):
    content = serializer.Field()
    success = serializer.Field()
    status_code = serializer.Field()
    redirect_location = serializer.Field()


class MailOutboxSerializer(serializer.SerpySerializer):
    to = serializer.Field()
    bcc = serializer.Field()
    cc = serializer.Field()
    from_email = serializer.Field()
    reply_to = serializer.Field()
    subject = serializer.Field()
    body = serializer.Field()
    content_subtype = serializer.Field()



import inspect
import os
import json
from os import path
from pathlib import Path
from shutil import which
from typing import List, Type

from w.services.technical.filesystem_service import FilesystemService
import factory.random

from w.tests.builders.abstract_test_builder import AbstractTestBuilder


class TestCaseMixin:
    """Mixin to handle common test functionalities"""

    _fixture_dir = None
    _w_fixture_dir = None

    base_dir = None

    @classmethod
    def factory_boy_seed_random(cls, seed=1) -> None:
        """
        Seeds the shared random number generator to generate same data

        Args:
            seed (int): seed

        Returns:
            None
        """
        if seed is None:
            seed = 1
        factory.random.reseed_random(seed)

    @classmethod
    def setup_class(cls):
        cls.factory_boy_seed_random()

    @classmethod
    def teardown_class(cls):
        cls.clean_sandbox()

    @classmethod
    def set_base_dir(cls, base_dir):
        cls.base_dir = base_dir

    @classmethod
    def get_base_dir(cls):
        """ Get base dir from django settings by default """
        if cls.base_dir is None:
            from django.conf import settings

            cls.base_dir = str(settings.BASE_DIR)
        return cls.base_dir

    @classmethod
    def get_tmp_dir(cls):
        """ Get tmp test directory """
        return f"{cls.get_base_dir()}/.donotcommit_tmp"

    @classmethod
    def get_tmp_diff_cmd_filename(cls):
        return f"{cls.get_tmp_dir()}/.donotcommit_tmp_diff_cmd"

    @classmethod
    def get_test_fixtures_dir(cls) -> str:
        """
        Get fixtures directory

        Returns:
            str: fixtures directory
        """
        if not cls._fixture_dir:
            filename = cls._get_calling_filename()
            pos = filename.find("/tests")
            if not pos > 0:
                raise ValueError(f"Unable to determine root tests path from {filename}")
            cls._fixture_dir = f"{filename[:pos]}/tests/fixtures"

        return cls._fixture_dir

    @classmethod
    def get_resultsets_dir(cls) -> str:
        """
        Get result sets dir

        Returns:
            str: result sets directory
        """
        return f"{cls.get_test_fixtures_dir()}/resultsets"

    @classmethod
    def get_datasets_dir(cls, filename) -> str:
        """
        get data sets path

        Args:
            filename (str): related path of filename

        Returns:
            str: data sets directory
        """
        return f"{cls.get_test_fixtures_dir()}/datasets/{filename}"

    @classmethod
    def get_w_datasets_dir(cls, filename) -> str:
        """
        get w data sets path

        Args:
            filename (str): related path of filename

        Returns:
            str: common data sets directory
        """
        if not cls._w_fixture_dir:
            w_dir = os.path.dirname(os.path.realpath(__file__))
            pos = w_dir.find("/tests")
            if not pos > 0:
                raise ValueError("Unable to determine root common tests path")
            cls._w_fixture_dir = f"{w_dir[:pos]}/tests/fixtures"
        return f"{cls._w_fixture_dir}/datasets/{filename}"

    @classmethod
    def get_app_datasets_dir(cls, filename, app) -> str:
        from django.conf import settings

        root_dir = Path(settings.BASE_DIR).parent
        return f"{root_dir}/{app}/tests/fixtures/datasets/{filename}"

    @classmethod
    def _get_dataset(cls, dataset, dataset_filename):
        """
        Get fixture dataset
        Args:
            dataset(str): relative path to dataset.
                          If is json, load it else just read it

        Returns:
            mixed
        """
        filename, file_extension = os.path.splitext(dataset)
        with open(dataset_filename) as file:
            if file_extension == ".json":
                return json.load(file)
            return file.read()

    @classmethod
    def get_dataset(cls, dataset):
        """
        Get fixture dataset
        Args:
            dataset(str): relative path to dataset.
                          If is json, load it else just read it

        Returns:
            mixed
        """
        return cls._get_dataset(dataset, cls.get_datasets_dir(dataset))

    @classmethod
    def get_w_dataset(cls, dataset):
        """
        Get fixture from wedge dataset
        Args:
            dataset(str): relative path to dataset.
                          If is json, load it else just read it

        Returns:
            mixed
        """
        return cls._get_dataset(dataset, cls.get_w_datasets_dir(dataset))

    @classmethod
    def get_app_dataset(cls, dataset, app):
        """
        Get fixture from django app dataset
        Args:
           dataset(str): relative path to dataset.
                         If is json, load it else just read it
           app(str): django app

        Returns:
           mixed
        """
        return cls._get_dataset(dataset, cls.get_app_datasets_dir(dataset, app))

    @classmethod
    def _save_as_fixture(cls, fixture_filename, dataset):
        fixture_filename = Path(fixture_filename)
        new_json = (
            json.dumps(
                cls._convert_data(dataset),
                separators=(",", ": "),
                default=str,
                # disable non-ASCII characters escape with \uXXXX sequences
                ensure_ascii=False,
            )
            + "\n"
        )  # for respecting file empty last line convention

        if fixture_filename.exists():
            # check if we need to update it
            actual = fixture_filename.read_text()
            if actual == new_json:
                return None
        else:
            cls._check_file_can_be_created(fixture_filename)

        # save dataset
        fixture_filename.write_text(new_json)

    @classmethod
    def save_as_dataset(cls, relative_filename, dataset):
        fixture_filename = cls.get_datasets_dir(relative_filename)
        cls._save_as_fixture(fixture_filename, dataset)

    @classmethod
    def get_sandbox_dir(cls, filename=None) -> str:
        """
        Get sandbox dir for current test

        Returns:
            str
        """
        sandbox_path = f"{cls.get_test_fixtures_dir()}/sandbox/{cls.__name__}"
        if filename is None:
            return sandbox_path
        filename = f"{sandbox_path}/{filename}"
        FilesystemService.check_file_can_be_created(filename)
        return filename

    @classmethod
    def clean_sandbox(cls):
        """
        Clean sandbox
        """
        if not os.path.exists(cls.get_sandbox_dir()):
            os.makedirs(cls.get_sandbox_dir())
        FilesystemService.empty_dir(cls.get_sandbox_dir())

    @classmethod
    def get_mock_calls(cls, mock):
        """
        Get mock calls
        Args:
            mock(MagicMock|dict): mock or mock list

        Returns:
            list|dict
        """
        if isinstance(mock, dict):
            calls = {}
            for method, mock in mock.items():
                calls[method] = cls.get_mock_calls(mock)
            return calls

        calls = []
        if mock.call_count > 0:
            for call in mock.call_args_list:
                args, kwargs = call
                calls.append({"args": list(args), "kwargs": kwargs})
            return calls

        for call in mock.method_calls:
            method, args, kwargs = call
            calls.append({"method": method, "args": list(args), "kwargs": kwargs})

        return calls

    def assert_equals_resultset(self, actual, **kwargs):
        """
        assert result set equality

        Build resultset name from filename.
        Compare expected to resultset for tests method
        """

        # build resultset filename from caller filename
        calling_filename = kwargs.get("calling_filename", None)
        filename = self._get_calling_filename()
        calling_frame = self._get_calframe(calling_filename)
        method_name = self._get_calling_method_name(calling_frame)
        sub_dir = self._get_test_subdir(filename)
        filename_noext = path.basename(filename).replace(".py", "")
        resultset_filename = (
            f"{self.get_resultsets_dir()}/{sub_dir}/{filename_noext}/{method_name}.json"
        )
        self._check_file_can_be_created(resultset_filename)

        # get expected result sets from file
        if os.path.exists(resultset_filename):
            with open(resultset_filename) as json_file:
                expected = json_file.read()
        else:
            # create missing file
            with open(resultset_filename, "w") as f:
                json.dump({}, f)
                expected = None

        actual_json = (
            json.dumps(
                self._convert_data(actual),
                indent=4,
                separators=(",", ": "),
                default=str,
                # disable non-ASCII characters escape with \uXXXX sequences
                ensure_ascii=False,
            )
            + "\n"  # for respecting file empty last line convention
        )
        # assert
        try:
            assert expected == actual_json
        except AssertionError as e:
            working_dir = self.get_base_dir()
            tmp_dir = self.get_tmp_dir()

            # build fix result set
            tmp_actual_filename = (
                f"{tmp_dir}/{filename_noext}-{method_name}_ACTUAL.json"
            )
            self._check_file_can_be_created(tmp_actual_filename)

            with open(tmp_actual_filename, "w") as f:
                f.write(actual_json)

            with open(self.get_tmp_diff_cmd_filename(), "a+") as f:
                common_cmd = f"{resultset_filename} {tmp_actual_filename}\n"
                if which("charm") is not None:
                    # f.write(f"charm diff {common_cmd}")
                    os.system(f"charm diff {common_cmd}")
                elif which("pycharm-community") is not None:
                    f.write(f"pycharm-community diff {common_cmd}")
                elif which("meld") is not None:
                    f.write(f"meld {common_cmd}")
                elif which("code") is not None:
                    f.write(f"code -d {common_cmd}")

            print(f"\n\033[1;31m=== TEST {method_name} has failed !!")
            print(f"file: {filename.replace(working_dir, '')}")
            os.system(f"diff {resultset_filename} {tmp_actual_filename}")
            print(f"=== TEST {method_name}\n\033[0m")
            raise e

    @classmethod
    def show_test_failed_diff(cls):
        """
        Show diff errors on failed tests
        Add this to your conftest.py for automatic call

            @pytest.fixture(scope="session", autouse=True)
            def show_diff_errors(request):
                request.addfinalizer(TestCaseMixin.show_diff)

        """
        filename = TestCaseMixin.get_tmp_diff_cmd_filename()
        if os.path.isfile(filename):
            with open(filename) as f:
                list_cmd = f.read()
            os.remove(filename)
            for cmd in list_cmd.split("\n"):
                os.system(cmd)  # noqa

    @staticmethod
    def assert_file_exists(filename):
        assert os.path.exists(filename)

    @staticmethod
    def assert_file_not_exists(filename):
        assert not os.path.exists(filename)

    @classmethod
    def _get_test_subdir(cls, filename):
        """
        Get subdir after tests directory
        Args:
            filename:

        Returns:

        """
        working_dir = cls.get_base_dir()
        dir_names = path.dirname(filename.replace(working_dir, "")).split("/")

        # search tests dir
        for _ in range(0, len(dir_names)):
            dir_name = dir_names.pop(0)
            if dir_name == "tests":
                break

        return "/".join(dir_names)

    @classmethod
    def _get_calframe(cls, calling_filename=None) -> inspect.FrameInfo:
        """
        find calling frame from stack calls
        Returns:
            FrameInfo : frame
        """
        # find calling frame
        calframe = inspect.getouterframes(inspect.currentframe(), 2)
        test_filename = (
            cls._get_calling_filename() if not calling_filename else calling_filename
        )
        for frame in calframe:
            if frame.filename.find(test_filename) >= 0:
                return frame
        raise ValueError("Unable to determine calling frame")

    @classmethod
    def _get_calling_filename(cls) -> str:
        """
        Returns:
            str: calling filename
        """
        return inspect.getfile(cls)

    @staticmethod
    def _get_calling_method_name(frame) -> str:
        """
        Returns:
            str: calling method name
        """
        return frame[3]

    @staticmethod
    def _get_calling_lineno(frame) -> str:
        """
        Returns:
            str: calling line number
        """
        return frame.lineno

    @staticmethod
    def _check_file_can_be_created(filename) -> None:
        """
        create missing directories for filename
        Args:
            filename (str) : fullpath filename

        Returns:
            None
        """
        return FilesystemService.check_file_can_be_created(filename)

    @classmethod
    def _convert_data(cls, data):
        from w.services.technical.json_service import JsonService

        return JsonService.to_json_dumpable(data)

    @staticmethod
    def reset_sequence_builders(
        list_builders: List[Type[AbstractTestBuilder]], reset_pk_model=True
    ):
        # remove duplicate
        list_builders = list(dict.fromkeys(list_builders))
        for builder in list_builders:
            builder.reset_sequence(reset_pk_model=reset_pk_model)



from io import StringIO

import pytest
import yaml

from w.services.technical.filesystem_service import FilesystemService
from w.services.technical.yaml_service import YamlService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestYamlService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.data = {
            "a value": "a data",
            "a list": [1, 12, 13, "string"],
            "a dict": {
                "key": "value",
                "key2": ["value", "valuebis"],
                "key3": {"dictk": "value"},
            },
        }

    """
    dump_to_mem_file
    """

    def test_dump_to_mem_file_with_success_return_stringio(self):
        """ Ensure dump_to_mem_file succeed """
        with open(self.get_datasets_dir("yaml/example.yml")) as f:
            data = yaml.safe_load(f)
            actual = YamlService.dump_to_mem_file(data)
            assert isinstance(actual, StringIO)
            assert data == yaml.safe_load(actual)

    """
    dump_to_file
    """

    def test_dump_to_file_with_success_return_none(self):
        """ Ensure method succeed """
        filename = self.get_sandbox_dir("simple_dump.yml")
        assert YamlService.dump_to_file(self.data, filename) is None
        self.assert_file_exists(filename)
        self.assert_equals_resultset(FilesystemService.read_file(filename).split("\n"))

    """
    load
    """

    def test_load_with_file_not_exists_raise_runtime_error(self):
        """ Ensure load unknown file raise RuntimeError """
        with pytest.raises(RuntimeError, match="unknown.yml does not exists"):
            YamlService.load_from_file("unknown.yml")

    def test_load_with_success_return_dict(self):
        """ Ensure load succeed """
        self.assert_equals_resultset(
            YamlService.load_from_file(self.get_datasets_dir("yaml/example.yml"))
        )

    """
    check_is_valid
    """

    def test_check_is_valid_with_no_valid_yaml_raise_runtime_error(self):
        """ Ensure check invalid yaml raise RuntimeError """
        invalid_yaml = FilesystemService.read_file(
            self.get_datasets_dir("yaml/invalid_yaml.yml")
        )

        match = "Line 3, column 24: could not find expected ':'"
        with pytest.raises(RuntimeError, match=match):
            YamlService.check_is_valid(yaml_stream=invalid_yaml)

    def test_check_is_valid_with_success_return_dict(self):
        """ Ensure check valid yaml, load yaml """
        valid_yaml = FilesystemService.read_file(
            self.get_datasets_dir("yaml/example.yml")
        )

        self.assert_equals_resultset(YamlService.check_is_valid(yaml_stream=valid_yaml))

    def test_check_is_valid_with_invalid_date_raise_runtime_error(self):
        """ Ensure check invalid date raises RuntimeError """
        yaml_with_invalid_date = FilesystemService.read_file(
            self.get_datasets_dir("yaml/yaml_with_invalid_date.yml")
        )

        match = "month must be in 1..12"
        with pytest.raises(RuntimeError, match=match):
            YamlService.check_is_valid(yaml_stream=yaml_with_invalid_date)



from rest_framework.exceptions import APIException
from w.django.utils import _  # noqa


class ServiceUnavailable(APIException):
    status_code = 503
    default_detail = "Service temporarily unavailable, try again later."


class WError(Exception):
    """
    Base class for exceptions

    Attributes:
        message (str): Error message
        params (dict) : message string params
        code (int): code like HTTP status code
    """

    status_code = 400
    message = "message undefined"

    def __init__(self, message=None, params=None, code=None) -> None:
        if not message:
            message = self.message
        if not code:
            code = self.status_code
        params = params if params else {}
        self.message = _(message) % params
        self.code = code

    def __str__(self) -> str:
        return f"{self.code} - {self.message}"

    def get_message(self):
        return self.message

    def get_code(self):
        return self.code


class InvalidCredentialsError(WError):
    """
    Raised when credentials are invalid
    """

    status_code = 401
    message = "Invalid Credentials"


class NotFoundError(WError):
    """Raised when somethings is not found"""

    status_code = 404

    def __init__(self, message, params=None) -> None:
        super().__init__(message, params)


class AlreadyExistsError(WError):
    """Raised when somethings already exists"""

    status_code = 409

    def __init__(self, message, params=None) -> None:
        super().__init__(message, params)


class ValidationError(WError):
    """Raised when data validation failed"""

    status_code = 422
    message = "Unprocessable Entity"
    detail = None

    def __init__(self, detail=None, params=None) -> None:
        super().__init__(self.message, params)
        self.detail = detail


class AuthenticationTimeoutError(WError):
    """Raise when Authentification Timeout"""

    status_code = 403
    message = "Authentication Timeout"

    def __init__(self, message=None, params=None) -> None:
        super().__init__(message, params)


class AccessForbiddenError(WError):
    """Raise when access is forbidden"""

    status_code = 403
    message = "Access Forbidden"

    def __init__(self, message=None, params=None) -> None:
        super().__init__(message, params)



from django import urls
from django.utils.translation import gettext


def reverse(
    viewname, urlconf=None, args=None, kwargs=None, current_app=None, query_kwargs=None
):
    """ "
    Url reverse extending django.urls.reverse to handle query_params

    Usage:
        reverse(
            <url_name>,
            kwargs={'pk': 123},
            query_kwargs={'key':'value', 'k2': 'v2'}
        )
    """
    base_url = urls.reverse(
        viewname, urlconf=urlconf, args=args, kwargs=kwargs, current_app=current_app
    )

    if query_kwargs:
        from w.services.technical.url_service import UrlService

        return f"{base_url}?{UrlService.get_url_query_string(**query_kwargs)}"

    return base_url


def _(message):
    """
    deprecated translation shortcut => old bad ideas
    better to use: from django.utils.translation import gettext as _
    """
    return gettext(message)


def list_accepted_types(request):
    """ List accepted types of HttpRequest """
    return [a.split(";")[0] for a in request.META["HTTP_ACCEPT"].split(",")]


def is_request_accept_json(request):
    """ Check if request accept json response """
    accepted_type = list_accepted_types(request)
    return "application/json" in accepted_type or "text/javascript" in accepted_type


def is_request_accept_html(request):
    """ Check if request accept html response """
    accepted_type = list_accepted_types(request)
    return "text/html" in accepted_type


def is_request_accept_all(request):
    """ Check if request accept all response """
    accepted_type = list_accepted_types(request)
    return "*/*" in accepted_type



import pytest

from w.services.technical.csv_service import CsvService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestCsvService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.filename = cls.get_datasets_dir("csv/simple.csv")
        cls.mapping = {"colA": "columnA", "colB": "columnB", "colC": "columnC"}

    """
    is_csv
    """

    def test_is_csv_with_no_csv_file_return_false(self):
        """ Ensure method return False on no csv file """
        files = ["toto.txt", "csv.txt", "csv.csv.gz"]
        for file in files:
            assert CsvService.is_csv(file) is False

    def test_is_csv_with_success_return_true(self):
        """ Ensure method return True on csv file """
        files = ["toto.csv", "csv.csv"]
        for file in files:
            assert CsvService.is_csv(file) is True

    """
    load
    """

    def test_load_with_invalid_csv_raise_runtime_error(self):
        """ Ensure method raise runtime error if file is not csv """
        filename = self.get_datasets_dir("excel/Excel2007File.xlsx")
        match = "Excel2007File.xlsx is not a csv file"
        with pytest.raises(RuntimeError, match=match):
            CsvService.load(filename, self.mapping)

    def test_load_with_invalid_header_raise_runtime_error(self):
        """ Ensure method raise runtime error if header is false """
        mapping = self.mapping.copy()
        mapping["colWrongA"] = mapping.pop("colA")
        match = (
            "incorrect or missing header, expected 'colB;colC;colWrongA'"
            " got 'colA;colB;colC'"
        )
        with pytest.raises(RuntimeError, match=match):
            CsvService.load(self.filename, mapping)

    def test_load_with_success_return_list(self):
        """ Ensure method succeed """
        self.assert_equals_resultset(CsvService.load(self.filename, self.mapping))

    """
    dump
    """

    def test_dump_with_success_return_none(self):
        """ Ensure method succeed """
        self.clean_sandbox()
        filename = self.get_sandbox_dir("test_dump.csv")
        mapping = {"colA": "columnA", "colB": "columnB", "colC": "columnC"}
        rows = [
            {
                "colA": "row 1 data, columnA",
                "colB": "row 1 data; columnB",
                "colC": "row 1 data\tcolumnC éàç",
            },
            {
                "colA": "row 2 data, columnA",
                "colB": "row 2 data, columnB",
                "colC": "row 2 data, columnC",
            },
        ]
        CsvService.dump(filename, mapping, rows)
        self.assert_file_exists(filename)
        reverse_mapping = {v: k for k, v in mapping.items()}
        self.assert_equals_resultset(CsvService.load(filename, reverse_mapping))



from w.django.tests.factories.auth_factories import UserFactory
from w.tests.builders.abstract_test_builder import AbstractTestBuilder


class UserTestBuilder(AbstractTestBuilder):
    _factories = {"user": UserFactory}
    _main_factory_key = "user"

    def build(self, **attrs):
        password = attrs.pop("password", None)
        user = super().build(**attrs)

        if password:
            user.set_password(password)
            user.save()

        return user



import factory
from w.django.models import Reference


class ReferenceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Reference

    code = factory.lazy_attribute(lambda o: o.label.upper())
    label = factory.Sequence(lambda n: f"reference_{n}")



import json

from django.db import models


# noinspection PyMethodMayBeStatic
class JsonField(models.TextField):
    def from_db_value(self, value, expression, connection):
        """ convert string from db to python dictionary """
        if value is None:
            return None
        return json.loads(value)

    def to_python(self, value):
        """ convert model input value to python dictionary """
        return None if value is None else json.dumps(value)



import pytest

from w.services.technical.abstract_import_service import AbstractImportService
from w.tests.helpers import service_test_helper
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestAbstractImportService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.mock_abstract_methods = [
            {
                "service": AbstractImportService,
                "method_name": "_list_mapping_columns",
                "return_value": None,
            },
            {
                "service": AbstractImportService,
                "method_name": "_validate",
                "return_value": ("validated_datas", "validation_errors"),
            },
            {
                "service": AbstractImportService,
                "method_name": "_import",
                "return_value": None,
            },
            {
                "service": AbstractImportService,
                "method_name": "_get_import_report",
                "return_value": "import_report",
            },
        ]
        cls.csv_mapping = {"colA": "columnA", "colB": "columnB", "colC": "columnC"}
        cls.excel_mapping = {"header1": "colA", "header2": "colB", "header3": "colC"}

    """
    import_file
    """

    def test_import_file_with_unsupported_format_raise_runtime_exception(self):
        """ Ensure we cannot import unsupported file format """
        match = "Format file1.txt not supported"
        with pytest.raises(RuntimeError, match=match):
            AbstractImportService.import_file(
                self.get_datasets_dir("filesystem/dir1/file1.txt")
            )

    def test_import_file_with_csv_return_data(self):
        """ Ensure we can import csv file """
        self.mock_abstract_methods[0]["return_value"] = self.csv_mapping
        with service_test_helper.mock_services(self.mock_abstract_methods) as m:
            import_report = AbstractImportService.import_file(
                self.get_datasets_dir("csv/simple.csv")
            )
        assert import_report == "import_report"
        self.assert_equals_resultset(self.get_mock_calls(m))

    def test_import_file_with_excel_return_data(self):
        """ Ensure we can import csv file """
        self.mock_abstract_methods[0]["return_value"] = self.excel_mapping
        with service_test_helper.mock_services(self.mock_abstract_methods) as m:
            import_report = AbstractImportService.import_file(
                self.get_datasets_dir("excel/Excel2010.xlsx")
            )
        assert import_report == "import_report"
        self.assert_equals_resultset(self.get_mock_calls(m))

    def test_import_file_with_sheet_name_return_data(self):
        """ Ensure we can import excel file with specific sheet name """
        excel_mapping = {
            "header1_sheet2": "columnA",
            "header2_sheet2": "columnB",
            "header3_sheet2_with_space": "columnC",
        }

        self.mock_abstract_methods[0]["return_value"] = excel_mapping
        with service_test_helper.mock_services(self.mock_abstract_methods) as m:
            AbstractImportService.sheet_name = "Sheet_2"
            import_report = AbstractImportService.import_file(
                self.get_datasets_dir("excel/Excel2010.xlsx")
            )
        assert import_report == "import_report"
        self.assert_equals_resultset(self.get_mock_calls(m))



from w.serializers import serializer


class ReferenceSerializer(serializer.SerpySerializer):
    code = serializer.StrField()
    label = serializer.StrField()


class I18nReferenceSerializer(ReferenceSerializer):
    label = serializer.TranslateField()


class UserSerializer(serializer.SerpySerializer):
    id = serializer.IntField()
    username = serializer.Field()
    first_name = serializer.Field()
    last_name = serializer.Field()
    email = serializer.Field()
    is_active = serializer.BoolField()


class UserWithDateSerializer(UserSerializer):
    date_joined = serializer.DateField()



import pytest

from w.services.technical.models.request_response import RequestResponse
from w.services.technical.models.request_session import RequestSession
from w.services.technical.request_service import RequestService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestRequestService(TestCaseMixin):
    @staticmethod
    def fix_response_variable_content(response):
        response["_content"]["headers"]["User-Agent"] = "pytest-user_agent"
        response["_content"]["headers"]["X-Amzn-Trace-Id"] = "X-Amzn-Trace-Id"
        # remove all headers beginning by X-
        response["_content"]["headers"] = {
            k: v for k, v in response["_content"]["headers"].items() if k[:2] != "X-"
        }
        response["_content"]["origin"] = "pytest"
        return response

    """
    init_session
    """

    def test_init_session_with_invalid_url_raise_runtime_error(self):
        with pytest.raises(RuntimeError, match="Invalid url 'http//foobar.d'"):
            RequestService.init_session("http//foobar.d")

    def test_init_session_with_success_return_request_session(self):
        session = RequestService.init_session(
            "http://url.com",
            headers={"x-test": "true"},
            auth=("username", "password"),
            cookies="cookies-data",
        )
        assert isinstance(session, RequestSession)
        assert session.base_url == "http://url.com"
        assert session.request.auth == ("username", "password")
        assert session.request.headers == {"x-test": "true"}
        assert session.request.cookies == "cookies-data"

    """
    get
    """

    def test_get_failed_return_response(self):
        actual = RequestService.get("http://httpbin.org/post")
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original models"
        actual._headers = "some headers"
        actual._content = actual._content.upper()
        self.assert_equals_resultset(actual.__dict__)

    @pytest.mark.skip(reason="httpbin not working as expected")
    def test_get_with_redirect_return_response(self):
        actual = RequestService.get(
            "http://httpbin.org/redirect-to?url=%2Ftoto%2Ftiti&status_code=302",
            allow_redirects=False,
        )
        assert isinstance(actual, RequestResponse)
        assert actual.status_code == 302
        assert actual._headers
        assert actual.is_redirect
        actual._headers = "some headers"
        self.assert_equals_resultset(actual.__dict__)

    def test_get_success_return_response(self):
        actual = RequestService.get("http://httpbin.org/get")
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original datas"
        actual._headers = "some headers"
        actual = actual.__dict__
        actual = self.fix_response_variable_content(actual)
        self.assert_equals_resultset(actual)

    def test_get_with_parameters_return_response(self):
        actual = RequestService.get(
            "http://httpbin.org/get", params={"p1": "valueA", "p2": "valueB"}
        )
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original datas"
        actual._headers = "some headers"
        actual = actual.__dict__
        actual = self.fix_response_variable_content(actual)
        self.assert_equals_resultset(actual)

    def test_get_with_timeout_return_503(self):
        actual = RequestService.get("http://httpbin.org/delay/3", timeout=0.001)
        assert isinstance(actual, RequestResponse)
        self.assert_equals_resultset(actual.__dict__)

    def test_get_with_session_and_auth_success_return_response(self):
        session = RequestService.init_session(
            "http://httpbin.org/", auth=("anUser", "aPassword")
        )
        actual = RequestService.get("/basic-auth/anUser/aPassword", session)
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._headers = "some headers"
        actual._orig_content = "original datas"
        self.assert_equals_resultset(actual.__dict__)

    def test_get_with_session_and_timeout_return_503(self):
        session = RequestService.init_session("http://httpbin.org/", timeout=0.001)
        actual = RequestService.get("/delay/3", session)
        assert isinstance(actual, RequestResponse)
        self.assert_equals_resultset(actual.__dict__)

    def test_get_with_pdf_io_return_response(self):
        """ Ensure we can get pdf """
        actual = RequestService.get(
            "https://www.w3.org/WAI/ER/tests/xhtml/testfiles/resources/pdf/dummy.pdf"
        )
        assert isinstance(actual, RequestResponse)
        assert actual

    """
    post
    """

    def test_post_failed_return_405(self):
        actual = RequestService.post(
            "http://httpbin.org/get", {"d1": "valueA", "d2": "valueB"}
        )
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original datas"
        actual._headers = "some headers"
        actual._content = actual._content.upper()
        self.assert_equals_resultset(actual.__dict__)

    def test_post_success_return_response(self):
        actual = RequestService.post(
            "http://httpbin.org/post", {"d1": "valueA", "d2": "valueB"}
        )
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original datas"
        actual._headers = "some headers"
        actual = actual.__dict__
        actual = self.fix_response_variable_content(actual)
        self.assert_equals_resultset(actual)

    def test_post_with_timeout_return_503(self):
        actual = RequestService.post(
            "http://10.255.255.1/timeout",
            {"d1": "valueA", "d2": "valueB"},
            timeout=0.001,
        )
        assert isinstance(actual, RequestResponse)
        self.assert_equals_resultset(actual.__dict__)

    def test_post_with_session_success_return_response(self):
        session = RequestService.init_session("http://httpbin.org/")
        actual = RequestService.post(
            "/post", data={"d1": "valueA", "d2": "valueB"}, session=session
        )
        assert isinstance(actual, RequestResponse)
        assert actual._orig_content
        assert actual._headers
        actual._orig_content = "original datas"
        actual._headers = "some headers"
        actual = actual.__dict__
        actual = self.fix_response_variable_content(actual)
        self.assert_equals_resultset(actual)



from w.serializers import serializer


class SimpleSerializer(serializer.SerpySerializer):
    integer = serializer.IntField()
    string = serializer.StrField()
    date = serializer.DateField()


class InternalDependencyOneSerializer(serializer.SerpySerializer):
    id = serializer.Field()
    name = serializer.Field()


class InternalDependencyTwoSerializer(serializer.SerpySerializer):
    id = serializer.Field()
    name = serializer.Field()


class ExampleSerializer(serializer.SerpySerializer):
    id = serializer.Field()
    name = serializer.Field()
    internal_one = InternalDependencyOneSerializer()
    internal_two = InternalDependencyTwoSerializer()



from w.services.technical.url_service import UrlService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestUrlService(TestCaseMixin):

    """
    get_url_query_string
    """

    def test_get_url_query_string_with_success_return_str(self):
        """ Ensure method succeed """
        assert "param1=value1" == UrlService.get_url_query_string(param1="value1")
        assert "param1=value1&param2=154" == UrlService.get_url_query_string(
            param1="value1", param2=154
        )
        assert (
            "param1=value1+with+%40+%24+%C3%B9&param2=154"
            == UrlService.get_url_query_string(param1="value1 with @ $ ù", param2=154)
        )

    """
    resolve_absolute_url
    """

    def test_resolve_absolute_url_with_success_return_str(self):
        """ Ensure method succeed """
        cases = [
            ("http://baseurl.test", "/relative/url"),
            ("http://baseurl.test/", "/relative/url"),
            ("http://baseurl.test", "relative/url"),
            ("http://baseurl.test/", "relative/url"),
        ]
        for base, relative in cases:
            assert (
                "http://baseurl.test/relative/url"
                == UrlService.resolve_absolute_url(base, relative)
            )



from w.django.tests.django_testcase import DjangoTestCase
from w.services.technical.date_service import DateService
from w.tests.fixtures.datasets.builders.builder_with_auto_now import (
    BuilderWithAutoNow,
)
from w.tests.fixtures.datasets.builders.builder_with_internal_dependencies import (
    BuilderWithInternalDependencies,
    InternalDependencyOneTestBuilder,
)
from w.tests.fixtures.datasets.builders.builder_with_related_dependencies import (
    BuilderWithRelatedDependencies,
)
from w.tests.fixtures.datasets.builders.simple_builder import SimpleBuilder


class TestTestBuilder(DjangoTestCase):
    @classmethod
    def setUpTestData(cls):
        pass

    """
    simple_builder
    """

    def test_simple_builder_with_success(self):
        example = SimpleBuilder().build()
        example2 = SimpleBuilder().build(attribute_one="specific")
        assert example != example2
        assert example.attribute_one == "value1"
        assert example.attribute_two == "value2"
        assert example2.attribute_one == "specific"
        assert example2.attribute_two == "value2"

    """
    builder_with_internal_dependencies
    """

    def test_builder_with_internal_dependencies_with_success(self):
        actual = (
            BuilderWithInternalDependencies()
            .with_internal_one(name="name1")
            .with_internal_two(name="name2")
            .build()
        )
        assert actual.internal_one.name == "name1"
        assert actual.internal_two.name == "name2"

    def test_builder_with_internal_dependencies_with_reset_sequence(self):
        BuilderWithInternalDependencies.reset_sequence()
        actual = BuilderWithInternalDependencies().build()
        assert actual.name == "example_internal_name_1"
        assert actual.internal_one.name == "internal_one_name_1"
        assert actual.internal_two.name == "internal_two_name_1"

    def test_builder_with_internal_builder_with_success(self):
        actual = (
            BuilderWithInternalDependencies()
            .with_internal_one(
                builder=InternalDependencyOneTestBuilder().with_attrs(name="builder")
            )
            .with_internal_two(name="name2")
            .build()
        )
        assert actual.internal_one.name == "builder"
        assert actual.internal_two.name == "name2"

    """
    builder_with_related_dependencies
    """

    def test_builder_with_related_dependencies(self):
        BuilderWithInternalDependencies.reset_sequence()
        builder = BuilderWithRelatedDependencies()
        related_data = [{"name": "my_related_3"}, {"name": "my_related_4"}]
        actual = (
            builder.with_related_dependency(name="my_related_1")
            .with_related_dependency(name="my_related_2")
            .with_related_dependencies(related_data)
            .with_nb_related_dependencies(nb=2)
            .with_nb_related_dependencies(nb=2, name="same name")
            .build()
        )
        relateds = builder.get_related_dependency_built()
        assert len(relateds) == 8
        names = []
        for related in relateds:
            names.append(related.name)
            assert actual == related.example
        expected = [
            "my_related_1",
            "my_related_2",
            "my_related_3",
            "my_related_4",
            "related_name_4",
            "related_name_5",
            "same name",
            "same name",
        ]
        assert expected == names

    """
    builder_with_auto_now
    """

    def test_builder_with_auto_now_with_success(self):
        example = BuilderWithAutoNow().with_today_is("2018-05-05").build()
        example2 = (
            BuilderWithAutoNow()
            .with_today_is("2007-07-07")
            .build(attribute_one="specific")
        )
        assert example != example2
        assert example.attribute_one == "value1"
        assert DateService.to_mysql_date(example.attribute_date) == "2018-05-05"
        assert example2.attribute_one == "specific"
        assert DateService.to_mysql_date(example2.attribute_date) == "2007-07-07"

    """
    build_post_data
    """

    def test_build_post_data_with_simple_builder(self):
        actual = SimpleBuilder().build_post_data()
        self.assert_equals_resultset(actual)

    def test_build_post_data_with_internal_dependencies(self):
        BuilderWithInternalDependencies.reset_sequence(1)
        actual = BuilderWithInternalDependencies().build_post_data()
        self.assert_equals_resultset(actual)



import datetime
import errno
import os
from decimal import Decimal
from importlib import import_module

from arrow import Arrow

from w.services.technical.dict_service import DictService
from w.services.technical.string_service import StringService


def import_path(path_module):
    """
    Dynamically import module

    Args:
        path_module(str): module path with . (ex: 'dir.dir2.file.Class')

    Returns:
        module
    """
    package, attr = path_module.rsplit(".", 1)
    return getattr(import_module(package), attr)


def dict_keep_keys(d, keys):
    """
    Remove key not in keys from dictionary

    Args:
        d(dict): dictionary to clean
        keys(list): dictionary keys to keep

    Returns:
        dict: cleaned dictionary
    """
    return DictService.keep_keys(d, keys)


def dict_remove_keys(d, keys):
    """
    Remove key in keys from dictionary

    Args:
        d(dict): dictionary to clean
        keys(list): dictionary keys to remove

    Returns:
        dict: cleaned dictionary
    """
    return DictService.remove_keys(d, keys)


def get_dict_last_values(d):
    """
    get last entry value
    Args:
        d(dict): dictonnary

    Returns:
        mixed
    """
    return DictService.get_last_entry_value(d)


def to_json_dumpable(data):
    """
    Convert data to be json dumpable
    """
    if isinstance(data, tuple):
        data = list(data)

    if isinstance(data, dict):
        for k, v in data.items():
            data[k] = to_json_dumpable(v)
        return data

    if isinstance(data, list):
        for i in data:
            to_json_dumpable(i)
        return data

    if isinstance(data, Decimal):
        return float(data)

    if isinstance(data, Arrow):
        return data.format()

    elif isinstance(data, datetime.datetime) or isinstance(data, datetime.date):
        return data.strftime("%Y-%m-%d %H:%M:%S")

    return data


def check_file_can_be_created(filename) -> None:
    """
    create missing directories for filename

    Args:
        filename (str) : fullpath filename

    Returns:
        None
    """
    if not os.path.exists(filename):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise


def round_of_rating(number):
    """
    Round a number to the closest half integer.

    Args:
        number:

    Returns:
        >>> round_of_rating(1.3)
        1.5
        >>> round_of_rating(2.6)
        2.5
        >>> round_of_rating(3.0)
        3.0
        >>> round_of_rating(4.1)
        4.0
    """

    return round(number * 2) / 2


def strip_accents(txt):
    """ Remove string accents """
    return StringService.strip_accents(txt)



import json

from django.conf import settings

from rest_framework import status as drf_status
from rest_framework.test import APIClient

from rest_framework_api_key.models import APIKey
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework.authtoken.models import Token

from w import Debug
from w.django.tests.django_testcase import DjangoTestCase
from w.django.tests.factories.auth_factories import UserFactory
from w.django.utils import reverse


class ApiTestCase(DjangoTestCase):
    """ Class to handle common API test functionalities """

    show_json_response = False
    default_user = None
    default_token = None
    default_api_key = None
    client_class = APIClient

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.factory_boy_seed_random()

    def _get_user(self):
        if self.default_user is None:
            self.default_user = UserFactory(
                username="factory-boy",
                first_name="Factory",
                last_name="Boy",
                email="fb@factory.com",
            )
        return self.default_user

    def force_authenticate(self, user=None):
        """
        Force authenticate, if no user given, use a default one
        (created via factory_boy)

        Args:
            user (User): user to authenticate (optional)
        """
        if user is None:
            user = self._get_user()
        self.client.force_authenticate(user=user)

    @staticmethod
    def _get_jwt_access_token(user):
        jwt_token = RefreshToken.for_user(user)
        return str(jwt_token.access_token)

    @staticmethod
    def _get_token(user):
        if hasattr(settings, "SIMPLE_JWT"):
            jwt_token = RefreshToken.for_user(user)
            return str(jwt_token.access_token)
        token, created = Token.objects.get_or_create(user=user)
        return token

    def _get_token_header(self, user):
        header_token_keyword = "Bearer" if hasattr(settings, "SIMPLE_JWT") else "Token"
        if user:
            token = self._get_token(user)
            return f"{header_token_keyword} {token}"

        if self.default_token is None:
            self.default_token = self._get_token(self._get_user())

        return f"{header_token_keyword} {self.default_token}"

    def _add_api_key_header(self, params):
        API_KEY_CUSTOM_HEADER = (
            settings.API_KEY_CUSTOM_HEADER
            if hasattr(settings, "API_KEY_CUSTOM_HEADER")
            else None
        )
        if self.default_api_key is None:
            _, self.default_api_key = APIKey.objects.create_key(
                name="my-remote-service"
            )
        params[API_KEY_CUSTOM_HEADER] = self.default_api_key

    def set_client_params(self, params, auth_by_token=True, user=None):
        if auth_by_token:
            params["HTTP_AUTHORIZATION"] = self._get_token_header(user)

        self._add_api_key_header(params)

    def _client_method(self, method, url, data=None, auth_by_token=True, user=None):
        params = {"path": url, "data": data}
        self.set_client_params(params, auth_by_token, user)
        return getattr(self.client, method)(**params)

    def client_get(self, url, auth_by_token=True, user=None):
        return self._client_method("get", url, auth_by_token=auth_by_token, user=user)

    def client_post(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "post", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_put(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "put", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_delete(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "delete", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def client_patch(self, url, data=None, auth_by_token=True, user=None):
        return self._client_method(
            "patch", url, data=data, auth_by_token=auth_by_token, user=user
        )

    def _assert_no_registered_user_return_401(self, method, urlconf, url_params):
        response = getattr(self.client, method)(reverse(urlconf, kwargs=url_params))
        self.assert_rest_status_code(response, drf_status.HTTP_401_UNAUTHORIZED)

    def _assert_unauthorized_user_return_403(
        self, method, url_name, url_params, user=None, data=None
    ):
        params = {"url": reverse(url_name, kwargs=url_params), "user": user}
        if data is not None:
            params["data"] = data
        response = getattr(self, f"client_{method}")(**params)
        self.assert_rest_status_code(response, drf_status.HTTP_403_FORBIDDEN)

    def _assert_unknown_resource_return_404(self, method, urlconf, url_params):
        response = getattr(self, f"client_{method}")(
            reverse(urlconf, kwargs=url_params)
        )
        self.assert_rest_status_code(response, drf_status.HTTP_404_NOT_FOUND)

    def assert_get_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot get if no user authenticated """
        self._assert_no_registered_user_return_401("get", urlconf, url_params)

    def assert_get_with_unauthorized_user_return_403(
        self, user, urlconf, url_params=None
    ):
        """ Ensure we cannot get if user is not authorized """
        self._assert_unauthorized_user_return_403("get", user, urlconf, url_params)

    def assert_get_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot get if resource is unknown """
        self._assert_unknown_resource_return_404("get", urlconf, url_params)

    def assert_post_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot post if no user authenticated """
        self._assert_no_registered_user_return_401("post", urlconf, url_params)

    def assert_post_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot post if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "post", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_post_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot post if resource is unknown """
        self._assert_unknown_resource_return_404("post", urlconf, url_params)

    def assert_put_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot put if no user authenticated """
        self._assert_no_registered_user_return_401("put", urlconf, url_params)

    def assert_put_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot put if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "put", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_put_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot put if resource is unknown """
        self._assert_unknown_resource_return_404("put", urlconf, url_params)

    def assert_patch_with_no_registered_user_return_401(self, urlconf, url_params=None):
        """ Ensure we cannot patch if no user authenticated """
        self._assert_no_registered_user_return_401("patch", urlconf, url_params)

    def assert_patch_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot patch if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "patch", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_patch_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot patch if resource is unknown """
        self._assert_unknown_resource_return_404("patch", urlconf, url_params)

    def assert_delete_with_no_registered_user_return_401(
        self, urlconf, url_params=None
    ):
        """ Ensure we cannot delete if no user authenticated """
        self._assert_no_registered_user_return_401("delete", urlconf, url_params)

    def assert_delete_with_unauthorized_user_return_403(
        self, urlconf, url_params=None, user=None, data=None
    ):
        """ Ensure we cannot delete if user is not authorized """
        self._assert_unauthorized_user_return_403(
            "delete", urlconf, url_params=url_params, user=user, data=data
        )

    def assert_delete_with_unknown_resource_return_404(self, urlconf, url_params=None):
        """ Ensure we cannot delete if resource is unknown """
        self._assert_unknown_resource_return_404("delete", urlconf, url_params)

    def assert_get_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_post_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_201_CREATED)

    def assert_put_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_patch_response(self, response):
        self.assert_rest_response(response, status=drf_status.HTTP_200_OK)

    def assert_delete_response(self, response):
        self.assert_rest_status_code(response, status=drf_status.HTTP_204_NO_CONTENT)

    @staticmethod
    def assert_rest_status_code(response, status, post_err_msg=""):
        try:
            assert status == response.status_code, (
                f"expected status {status} == {response.status_code} "
                f"actual{post_err_msg}"
            )
        except Exception as e:
            Debug.s(json.loads(response.content))
            raise e

    def assert_rest_response(self, response, status):
        self.assert_rest_status_code(response, status)
        if self.show_json_response:
            Debug.s(response.content)
        self.assert_equals_resultset(json.loads(response.content))

    def assert_invalid_data_response(self, response):
        self.assert_rest_response(
            response, status=drf_status.HTTP_422_UNPROCESSABLE_ENTITY
        )

    @classmethod
    def save_as_cypress_fixtures(cls, fixture_path, dataset):
        """ Save dataset into cypress fixture repo """
        if settings.CYPRESS_FIXTURE_DIR is None:
            # ignore cypress fixture saving
            return None

        fixture_filename = f"{settings.CYPRESS_FIXTURE_DIR}/{fixture_path}"
        cls._save_as_fixture(fixture_filename, dataset)

    @classmethod
    def save_response_as_cypress_fixtures(cls, fixture_path, response):
        """ Save response content into cypress fixture repo """
        return cls.save_as_cypress_fixtures(fixture_path, json.loads(response.content))



from contextlib import contextmanager
from dataclasses import dataclass
from unittest.mock import patch

from w.tests.helpers import date_test_helper


@dataclass
class FakeRefreshToken:
    access_token: str
    refresh_token: str

    def __str__(self) -> str:
        return self.refresh_token


@contextmanager
def with_refresh_token(
    access_token="mock-access-token", refresh_token="mock-refresh-token"
):
    target = "rest_framework_simplejwt.serializers.TokenObtainPairSerializer.get_token"
    token = FakeRefreshToken(access_token, refresh_token)
    with date_test_helper.today_is("2020-01-01 11:11:11"):
        with patch(target, return_value=token) as m:
            yield m



import csv
from pathlib import Path

from w.services.abstract_service import AbstractService
from w.services.technical.filesystem_service import FilesystemService


class CsvService(AbstractService):
    @classmethod
    def is_csv(cls, filename):
        """ Check if filename is csv (from extension) """
        return FilesystemService.has_suffix(filename, ".csv")

    @classmethod
    def check_is_csv(cls, filename) -> Path:
        """
        Check if filename is csv (from extension) or raise RuntimeError
        """
        filename = FilesystemService.get_path(filename)
        if cls.is_csv(filename):
            return filename
        raise RuntimeError(f"{filename.name} is not a csv file")

    @classmethod
    def load(cls, csv_filename, mapping_columns) -> list:
        """
        Load csv file to list

        Args:
            csv_filename(str|Path): csv file
            mapping_columns(dict): csv header columns mapping to wanted attributes

        Returns:
            list: [{"<mapping name>": < row col value>, ...}, ...]

        Raises:
            RuntimeError :
                - filename is not csv
                - filename does not exists
                - incorrect or missing header
        """
        csv_filename = cls.check_is_csv(csv_filename)
        FilesystemService.check_file_exists(csv_filename)
        reader = csv.reader(csv_filename.open())
        required_headers = list(mapping_columns.keys())
        headers = [c for c in next(reader)]
        if required_headers != headers:
            raise RuntimeError(
                f"incorrect or missing header, expected '{';'.join(required_headers)}' "
                f"got '{';'.join(headers)}'"
            )
        return [
            {mapping_columns[headers[i]]: value for i, value in enumerate(row)}
            for row in reader
        ]

    @classmethod
    def dump(cls, csv_filename, mapping_columns, rows):
        """
        Dump rows into csv

        Args:
            csv_filename(str|Path): csv filename
            mapping_columns(dict): attributes to csv header columns mapping
            rows([dict]): list of rows [{"<mapping name>": <row col value>, ...}, ...]
        """
        with open(csv_filename, "w") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=mapping_columns.values())
            writer.writeheader()
            for row in rows:
                csv_row = {
                    mapping_columns[k]: v
                    for k, v in row.items()
                    if k in mapping_columns
                }
                writer.writerow(csv_row)



from abc import ABC


class AbstractService(ABC):
    """
    Ensure only one instance can be created, kind of singleton pattern
    """

    def __new__(cls, *args, **kwargs):
        raise RuntimeError("%s should not be instantiated" % cls)

    @classmethod
    def clear(cls):
        pass  # is like an abstract class method

    @classmethod
    def _is_initialized(cls):
        return True

    @classmethod
    def _check_is_initialized(cls):
        """
        Check service is initialized

        Returns:
            None

        Raises:
            RuntimeError
        """
        if cls._is_initialized():
            return None

        raise RuntimeError(f"Service {cls.__name__} must be initialized first")

    @classmethod
    def get_patch_target(cls, method_name=None, relative=False):
        """
        Get patch target to easily patch/mock the service
        """
        import inspect

        module = None
        for k, v in inspect.getmembers(cls):
            if k == "__module__":
                module = v
                break

        target = f"{module}" if relative else f"{module}.{cls.__name__}"
        if method_name:
            return f"{target}.{method_name}"
        return target



from w.services.technical.string_service import StringService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestExcelService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()

    """
    strip_accents
    """

    def test_strip_accents_with_success_return_cleaned_txt(self):
        """ Ensure method succeeds """
        txt = "àéèïôù"
        actual = StringService.strip_accents(txt)
        self.assert_equals_resultset(actual)

    """
    clean
    """

    def test_clean_with_no_replacement_return_cleaned_txt(self):
        """ Ensure method succeeds """
        txt = "  tHis Is à str-iNg 2 clé@N :  "
        actual = StringService.clean(txt)
        self.assert_equals_resultset(actual)

    def test_clean_with_success_return_cleaned_txt(self):
        """ Ensure method succeeds """
        txt = "  tHis Is à str-iNg 2 clé@N :  "
        actual = StringService.clean(txt, {"2": "to", "@": "a", " :": ".", "-": ""})
        self.assert_equals_resultset(actual)



import inspect
import json
import os

from django.db.models import QuerySet
from var_dump import var_dump
from sys import exit

from w import utils


class Debug:
    @staticmethod
    def d(var):
        """Prints HTML human-readable information about a variable"""
        Debug.dump(var, False)

    @staticmethod
    def dd(var):
        """Prints (and DIE) HTML human-readable information about a variable"""
        Debug.dump(var)

    @staticmethod
    def s(var):
        """Prints human-readable information about a variable"""
        Debug.dump(var, False, False)

    @staticmethod
    def sd(var):
        """Prints (and DIE) human-readable information about a variable"""
        Debug.dump(var, True, False)

    @staticmethod
    def j(var):
        """Prints human-readable json information about a variable"""
        Debug.dump(var, False, False, True)

    @staticmethod
    def jd(var):
        """Prints (and DIE) human-readable json information about a variable"""
        Debug.dump(var, True, False, True)

    @classmethod
    def dump(cls, var, die=True, html=True, into_json=False):
        """Prints human-readable information about a variable

        Args:
            var (mixed): variable to dump
            die (bool) : exit script. Defaults True
            html (bool): dump as HTML. Defautls True
            into_json (bool): dump as json. Defautls False
        """
        # build resultset filename from caller filename
        caller = inspect.getouterframes(inspect.currentframe(), 2)[2]
        working_dir = os.getcwd()
        filename = caller.filename.replace(working_dir, "")

        if into_json:
            var = json.dumps(utils.to_json_dumpable(var), indent=4)

        print("\n\n=== DEBUG FROM {}:{}\n".format(filename, caller.lineno))
        if html:
            print("<pre>")

        if not isinstance(var, QuerySet):
            var_dump(var)
        else:
            print(var)

        if html:
            print("</pre>")

        print("\n=== FIN DEBUG {}:{}\n".format(filename, caller.lineno))

        if die:
            exit()



import pytest
from django.template import TemplateDoesNotExist

from w.services.technical.pdf_service import PdfService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestPdfService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.filename = "test_pdf_service.pdf"

    def test_create_pdf_with_empty_content_return_none(self):
        """
        Ensure pdf service return None with empty content param
        """
        self.clean_sandbox()
        params = {
            "filename": self.get_sandbox_dir(filename=self.filename),
            "content": {},
        }
        pdf = PdfService.write_file(**params)
        assert pdf is None
        self.assert_file_not_exists(params["filename"])

    def test_create_pdf_with_unknown_template_raise_exception(self):
        """
        Ensure pdf service raise exception with unknown template path
        """
        self.clean_sandbox()
        params = {
            "filename": self.get_sandbox_dir(filename=self.filename),
            "content": {"template_name": "unknown_template.html", "context": ""},
        }
        match = "unknown_template.html"
        with pytest.raises(TemplateDoesNotExist, match=match):
            PdfService.write_file(**params)
        self.assert_file_not_exists(params["filename"])

    def test_create_pdf_with_full_template_file_return_bytes(self):
        """
        Ensure pdf service return a pdf with html template
        """
        self.clean_sandbox()
        pdf_context = {"name": "test_name"}
        params = {
            "filename": self.get_sandbox_dir(filename=self.filename),
            "content": {
                "template_name": "pdf/test_pdf_service.html",
                "context": pdf_context,
            },
        }
        pdf = PdfService.write_file(**params)
        assert pdf is not None and isinstance(pdf, bytes)
        self.assert_file_exists(params["filename"])

    def test_create_pdf_with_empty_context_return_bytes(self):
        """
        Ensure pdf service return a pdf with html template and empty context
        """
        self.clean_sandbox()
        params = {
            "filename": self.get_sandbox_dir(filename=self.filename),
            "content": {"template_name": "pdf/test_pdf_service.html", "context": {}},
        }
        pdf = PdfService.write_file(**params)
        assert pdf is not None and isinstance(pdf, bytes)
        self.assert_file_exists(params["filename"])

    def test_create_pdf_with_html_string_return_bytes(self):
        """
        Ensure pdf service return a pdf with html string
        """
        self.clean_sandbox()
        params = {
            "filename": self.get_sandbox_dir(filename=self.filename),
            "content": "<h1>test</h1>",
        }
        pdf = PdfService.write_file(**params)
        assert pdf is not None and isinstance(pdf, bytes)
        self.assert_file_exists(params["filename"])



from django.http import HttpResponse, HttpResponseRedirect
from django.http.response import HttpResponseBase

from w.django.fixtures.factory_boy import UserFactory
from w.django.tests.django_testcase import DjangoTestCase
from w.django.utils import reverse


class ViewTestCase(DjangoTestCase):
    """ Class to handle common View test functionalities """

    def _get_user(self):
        if self.default_user is None:
            self.default_user = UserFactory(
                username="factory-boy",
                first_name="Factory",
                last_name="Boy",
                email="fb@factory.com",
            )
        return self.default_user

    def force_login(self, user=None):
        """
        Force login, if no user given, use a default one
        (created via factory_boy)

        Args:
            user (User): user to authenticate (optional)
        """
        if user is None:
            user = self._get_user()
        self.client.force_login(user=user)

    def _client_method(self, method, url, force_login, data=None, user=None, **extra):
        params = {"path": url, "data": data}
        if force_login:
            self.force_login(user)
        return getattr(self.client, method)(**{**params, **extra})

    def client_get(self, url, user=None, force_login=True):
        return self._client_method("get", url, force_login, user=user)

    def client_post(self, url, data=None, user=None, force_login=True):
        return self._client_method("post", url, force_login, data=data, user=user)

    def client_ajax_get(self, url, user=None, force_login=True):
        return self._client_method(
            "get",
            url,
            force_login,
            user=user,
            HTTP_ACCEPT="text/html, */*; q=0.01",
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )

    def client_ajax_post(self, url, data=None, user=None, force_login=True):
        return self._client_method(
            "post",
            url,
            force_login,
            data=data,
            user=user,
            HTTP_ACCEPT="text/html, */*; q=0.01",
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )

    @staticmethod
    def assert_response_status(response: HttpResponse, status_code=200):
        """ Assert response status """
        assert response.status_code == status_code

    @staticmethod
    def assert_redirect(response: HttpResponseBase, redirect_url=None):
        """ Assert response is redirect with url """
        assert isinstance(response, HttpResponseRedirect)
        assert response.status_code == HttpResponseRedirect.status_code
        if redirect_url is not None:
            assert response.url == redirect_url, f"{response.url} != {redirect_url}"

    def assert_redirect_login(self, response, redirect_url):
        self.assert_redirect(response, f"{reverse('login')}?next={redirect_url}")

    def assert_get_with_unauthenticated_user_redirect_login(self, url):
        """ Assert get with unauthenticated user is redirected to login """
        response = self.client_get(url, force_login=False)
        self.assert_redirect_login(response, url)

    def assert_post_with_unauthenticated_user_redirect_login(self, url):
        """ Assert post with unauthenticated user is redirected to login """
        response = self.client_post(url, data={"param": "value"}, force_login=False)
        self.assert_redirect_login(response, url)



from w.services.technical.dict_service import DictService
from w.tests.mixins.testcase_mixin import TestCaseMixin


class TestDictService(TestCaseMixin):
    @classmethod
    def setup_class(cls):
        cls.data = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3",
            "key4": "value4",
        }

    """
    keep_keys
    """

    def test_keep_keys_with_no_key_found_return_dict(self):
        """ Ensure method succeed even if all keys to keep do not exist """
        actual = DictService.keep_keys(self.data, ["unknown1", "unknown2"])
        assert actual == {}

    def test_keep_keys_with_success_return_dict(self):
        """ Ensure method succeed """
        actual = DictService.keep_keys(
            self.data, ["value1", "key2", "unknown2", "key4"]
        )
        self.assert_equals_resultset(actual)

    """
    remove_keys
    """

    def test_remove_keys_with_no_key_found_return_dict(self):
        """ Ensure method succeed even if all keys to remove do not exist """
        actual = DictService.remove_keys(self.data, ["unknown1", "unknown2"])
        self.assert_equals_resultset(actual)

    def test_remove_keys_with_success_return_dict(self):
        """ Ensure method succeed """
        actual = DictService.remove_keys(
            self.data, ["value1", "key2", "unknown2", "key4"]
        )
        self.assert_equals_resultset(actual)

    """
    get_last_entry_value
    """

    def test_get_last_entry_value_with_success_return_dict(self):
        """ Ensure method succeed """

        actual = DictService.get_last_entry_value(self.data)
        self.assert_equals_resultset(actual)



from django.db import models


class AbstractCreatedUpdatedModel(models.Model):
    modified = models.DateTimeField(auto_now_add=True)
    created = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True



from django.db import models


class Example(models.Model):
    attribute_one = models.CharField(max_length=50)
    attribute_two = models.CharField(max_length=50)

    def __str__(self):
        return f"{self.attribute_one}/{self.attribute_two}"


class FkOne(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return f"{self.name}"


class FkTwo(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return f"{self.name}"


class ExampleWithFks(models.Model):
    name = models.CharField(max_length=50)
    fk_one = models.ForeignKey(FkOne, on_delete=models.CASCADE)
    fk_two = models.ForeignKey(FkTwo, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.name} ({self.fk_one}/{self.fk_two})"


class InternalDependencyOne(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return f"{self.name}"


class InternalDependencyTwo(models.Model):
    name = models.CharField(max_length=50)

    def __str__(self):
        return f"{self.name}"


class ExampleWithInternalDependencies(models.Model):
    name = models.CharField(max_length=50)
    internal_one = models.ForeignKey(InternalDependencyOne, on_delete=models.CASCADE)
    internal_two = models.ForeignKey(InternalDependencyTwo, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.name} ({self.internal_one}/{self.internal_two})"



from django.utils.http import urlencode

from w.services.abstract_service import AbstractService


class UrlService(AbstractService):
    @staticmethod
    def get_url_query_string(**query_params):
        return f"{urlencode(query_params, doseq=True)}"

    @staticmethod
    def resolve_absolute_url(base_url, relative_url):
        if base_url[-1:] == "/":
            # remove ending /
            base_url = base_url[:-1]

        if relative_url[:1] != "/":
            # add starting /
            relative_url = f"/{relative_url}"

        return f"{base_url}{relative_url}"



import logging

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import exception_handler as drf_exception_handler

from w import exceptions

logger = logging.getLogger(__name__)


def exception_handler(exc, context):
    """

    Args:
        exc(Exception|CommonError): exception
        context:

    Returns:

    """
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = drf_exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if response is None:
        if isinstance(exc, exceptions.WError):
            if isinstance(exc, exceptions.ValidationError):
                return Response(exc.detail, status=status.HTTP_422_UNPROCESSABLE_ENTITY)
            return Response(exc.get_message(), status=exc.get_code())

        logger.error(str(exc))
        return Response(str(exc), status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    return response



import copy
import datetime

import arrow
from arrow import Arrow

from w.services.abstract_service import AbstractService


class DateService(AbstractService):
    """Service Date"""

    # day number
    MON = 0
    TUE = 1
    WED = 2
    THU = 3
    FRI = 4
    SAT = 5
    SUN = 6

    # timezone if none
    _timezone = None

    @classmethod
    def clear(cls):
        super().clear()
        cls._timezone = None

    @classmethod
    def _get(cls, d, default_now=True) -> Arrow:
        """
        Convert to arrow if necessary
        Args:
            d(None|str|Arrow|datetime.datetime|datetime.date): date, today if None

        Returns:
            Arrow
        """
        if default_now and d is None:
            return arrow.utcnow()

        if isinstance(d, Arrow):
            return d
        return arrow.get(d)

    @classmethod
    def _format(cls, d, date_format) -> str:
        """
        Format with timezone if set

        Args:
            d(None|str|Arrow|datetime.datetime|datetime.date): date, today if None
            date_format(string): string format

        Returns:
            str
        """
        if cls._timezone:
            return cls._get(d).to(cls._timezone).format(date_format)
        return cls._get(d).format(date_format)

    @classmethod
    def set_timezone(cls, timezone):
        """
        Define default timezone

        Args:
            timezone(str):
        """
        cls._timezone = timezone

    @staticmethod
    def get(d) -> Arrow:
        """
        Get Arrow date from d
        Args:
            d(str|datetime.datetime|datetime.date)

        Returns:
            Arrow
        """
        return arrow.get(d)

    @staticmethod
    def now() -> Arrow:
        """
        Get now date
        Returns:
            Arrow
        """
        return arrow.utcnow()

    @classmethod
    def previous_weekday(cls, d, seeking_day) -> Arrow:
        """
        Determine the first previous weekday from a date

        Args:
            d (str|Arrow|datetime.datetime|datetime.date): arrow date to search from
            seeking_day (int):  seeking day num (use Date.MON, Date.TUE ...)

        Returns:
            Arrow: date to previous seeking day
        """
        d = cls._get(d, default_now=False)
        current_day = d.weekday()
        previous_date = copy.copy(d)

        if current_day == seeking_day:
            return d
        if current_day > seeking_day:
            nb_day = seeking_day - current_day
        else:
            nb_day = seeking_day - current_day - 7

        return previous_date.shift(days=nb_day)

    @classmethod
    def previous_month(cls, d=None) -> Arrow:
        """
        Determine the previous weekday from a date

        Args:
            d(None|str|Arrow|datetime.datetime|datetime.date): today if not set

        Returns:
            Arrow: Last day of previous month
        """
        return cls._get(d).replace(day=1).shift(days=-1)

    @classmethod
    def next_weekday(cls, d, seeking_day) -> Arrow:
        """
        Determine the first next weekday from a date

        Args:
            d (str|Arrow|datetime.datetime|datetime.date): arrow date to search from
            seeking_day (int):  seeking day (use DateService.MON, DateService.TUE ...)

        Returns:
            Arrow: date to previous seeking day
        """
        d = cls._get(d, default_now=False)
        current_day = d.weekday()
        next_date = copy.copy(d)

        if current_day == seeking_day:
            return d
        if current_day < seeking_day:
            nb_day = seeking_day - current_day
        else:
            nb_day = seeking_day - current_day + 7

        return next_date.shift(days=nb_day)

    @classmethod
    def get_weekday_date(cls, d, seeking_day) -> Arrow:
        """
        Determine seeking_day date from a date in same week

        Args:
            d(str|Arrow|datetime.datetime|datetime.date): date to search from
            seeking_day (int):  seeking day (use DateService.MON, DateService.TUE ...)

        Returns:
            Arrow: date of seeking day
        """

        d = cls._get(d, default_now=False)

        current_day = d.weekday()
        if current_day == seeking_day:
            return d

        return d.shift(days=(seeking_day - current_day))

    @classmethod
    def to_mysql_date(cls, d=None) -> str:
        """
        Format date to mysql format string

        Args:
            d(None|str|Arrow|datetime.datetime|datetime.date): today if not set

        Returns:
            str: mysql date
        """
        return cls._format(d, "YYYY-MM-DD")

    @classmethod
    def to_mysql_datetime(cls, d=None) -> str:
        """
        Format datetime to mysql format string

        Args:
            d(None|str|Arrow|datetime.datetime|datetime.date): today if not set

        Returns:
            str: mysql date
        """
        return cls._format(d, "YYYY-MM-DD HH:mm:ss.SSSSSS")

    @classmethod
    def to_str(cls, date_format, d=None) -> str:
        """
        Format date to str to a format string

        Args:
            date_format(string): format string
            d(None|str|Arrow|datetime.datetime|datetime.date): today if not set

        Returns:
            str: mysql date
        """
        return cls._format(d, date_format)

    @staticmethod
    def week_number(d=None) -> int:
        """
        Get week number from date

        Args:
            d(Arrow|datetime.datetime|datetime.date):

        Returns:
            int
        """
        if d is None:
            d = datetime.datetime.today()
        elif isinstance(d, Arrow):
            d = d.datetime

        return d.isocalendar()[1]

    @classmethod
    def to_date(cls, d=None) -> datetime.date:
        """
        Convert date to datetime.date

        Args:
            d(None|str|Arrow): today if not set

        Returns:

        """
        return cls._get(d).date()

    @classmethod
    def to_datetime(cls, d=None) -> datetime.date:
        """
        Convert date to datetime.datetime.

        Args:
            d(None|str|Arrow): today if not set

        Returns:

        """
        d = cls._get(d)
        if cls._timezone:
            return d.to(cls._timezone).datetime
        return d.naive

    @staticmethod
    def _check_date_start_before_end(start, end):
        """
        Check if start if before end
        Args:
            start:
            end:

        Returns:

        """
        if end < start:
            raise RuntimeError("start must be before end")

    @classmethod
    def is_between(cls, d, start, end) -> bool:
        """
        Check if date is between start and end
        Args:
            d(str|Arrow|datetime.datetime|datetime.date): date to check
            start(str|Arrow|datetime.datetime|datetime.date): start date
            end(str|Arrow|datetime.datetime|datetime.date): end date

        Returns:
            bool
        """
        d = cls._get(d, default_now=False)
        start = cls._get(start)
        end = cls._get(end)

        cls._check_date_start_before_end(start, end)

        return start <= d <= end

    @classmethod
    def range_day(cls, start, end, **kwargs):
        """
        Returns an iterator of :class:`Arrow <arrow.arrow.Arrow>` objects, representing
        points in time between two inputs.

        Args:
            start(str|Arrow|datetime.datetime|datetime.date): start date
            end(str|Arrow|datetime.datetime|datetime.date): end date
            **kwargs: only_week_end : retrieve only weekend day (default False)

        Returns:
            list(Arrow)
        """
        exclude = []

        if kwargs.pop("only_week_end", False):
            exclude = range(cls.MON, cls.SAT)

        start = cls._get(start, default_now=False)
        end = cls._get(end, default_now=False)

        cls._check_date_start_before_end(start, end)

        result = []
        for d in arrow.Arrow.range("day", start, end):
            if d.weekday() not in exclude:
                result.append(d)
        return result

    @classmethod
    def range_month(cls, start, end):
        """
        Returns an iterator of :class:`Arrow <arrow.arrow.Arrow>` objects, representing
        points in time between two inputs.

        Args:
            start(str|Arrow|datetime.datetime|datetime.date): start date
            end(str|Arrow|datetime.datetime|datetime.date): end date

        Returns:
            list(Arrow)
        """
        start = cls._get(start, default_now=False)
        end = cls._get(end, default_now=False)

        cls._check_date_start_before_end(start, end)
        return arrow.Arrow.range("month", start, end)

    @classmethod
    def is_week_end(cls, d=None) -> bool:
        """
        Check is date is week end

        Args:
           d(str|Arrow|datetime.datetime|datetime.date): date to check, default today

        Returns:
            bool
        """
        d = cls._get(d)
        return d.weekday() in [DateService.SAT, DateService.SUN]

    @classmethod
    def add_days(cls, d, nb, exclude_we=False) -> Arrow:
        """
        Retrieve date by adding nb days. Optionally ignore week ends
        Args:
            d(str|Arrow|datetime.datetime|datetime.date): date
            nb(int): nb days to add
            exclude_we(bool): ignore week ends

        Returns:
            Arrow
        """
        d = cls._get(d)
        while nb > 0:
            d = d.shift(days=+1)
            if not exclude_we or not DateService.is_week_end(d):
                nb -= 1
        return d

    @classmethod
    def nb_days(cls, start, end, exclude_we=False):
        """
        Nb days between 2 dates
        Args:
            start(str|Arrow|datetime.datetime|datetime.date): start date
            end(str|Arrow|datetime.datetime|datetime.date): end date
            exclude_we(bool): ignore week ends

        Returns:

        """
        start = cls._get(start, default_now=False)
        end = cls._get(end, default_now=False)

        if not exclude_we:
            delta = end - start
            return delta.days

        nb = 0
        arrow_range = (
            arrow.Arrow.range("day", start, end)
            if start <= end
            else arrow.Arrow.range("day", end, start)
        )
        for d in arrow_range:
            if not cls.is_week_end(d):
                nb += 1

        if start <= end:
            return nb

        return nb * -1



from django.utils import timezone


class Example:
    def __init__(self, attribute_one, attribute_two):
        self.attribute_one = attribute_one
        self.attribute_two = attribute_two

    def __str__(self):
        return f"{self.attribute_one}/{self.attribute_two}"


class InternalDependencyOne:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name}"


class InternalDependencyTwo:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name}"


class ExampleWithInternalDependencies:
    def __init__(
        self,
        name,
        internal_one: InternalDependencyOne,
        internal_two: InternalDependencyTwo,
    ):
        self.name = name
        self.internal_one = internal_one
        self.internal_two = internal_two

    def __str__(self):
        return f"{self.name}"


class RelatedDependency:
    def __init__(self, name, example: ExampleWithInternalDependencies):
        self.name = name
        self.example = example

    def __str__(self):
        return f"{self.name}"


class ExampleWithAutoNow:
    def __init__(self, attribute_one):
        self.attribute_one = attribute_one
        self.attribute_date = timezone.now()

    def __str__(self):
        return f"{self.attribute_one}/{self.attribute_date}"


class FakeDjangoModel:
    def __init__(self, id, attribute_one):
        self.id = id
        self.pk = id
        self.attribute_one = attribute_one


class FakeDjangoModelWithForeignKey:
    def __init__(self, id, name, fk_one: FakeDjangoModel):
        self.id = id
        self.pk = id
        self.name = name
        self.fk_one = fk_one



from w.tests.builders.abstract_test_builder import AbstractTestBuilder
from w.tests.fixtures.datasets.builders.factory_boy import (
    ExampleWithInternalDependenciesFactory,
    InternalDependencyOneFactory,
    InternalDependencyTwoFactory,
)


class InternalDependencyOneTestBuilder(AbstractTestBuilder):
    _factories = {
        "internal_one": InternalDependencyOneFactory,
    }
    _main_factory_key = "internal_one"


class BuilderWithInternalDependencies(AbstractTestBuilder):
    _factories = {
        "example": ExampleWithInternalDependenciesFactory,
        "internal_one": InternalDependencyOneFactory,
        "internal_two": InternalDependencyTwoFactory,
    }
    _main_factory_key = "example"

    def with_internal_one(self, builder=None, **attrs):
        if builder:
            self._set_foreign_key_builder("internal_one", builder)
        else:
            self._set_foreign_key_factory_attrs("internal_one", attrs)
        return self

    def with_internal_two(self, **attrs):
        self._set_foreign_key_factory_attrs("internal_two", attrs)
        return self
